/*! \file Header.h
\brief Header for all procedures

All necessary CGAL classes are included and typedefs are set. Routine inline functions are regrouped here.

\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re
*/

#ifndef MY_HEADER_H
#define MY_HEADER_H

#include <omp.h>
#include <boost/math/constants/constants.hpp>
#include "mex.h" 
#include <CGAL/Simple_cartesian.h>
#include <CGAL/Polyhedron_3.h>
#include <CGAL/Bbox_3.h>
#include <CGAL/Timer.h>

/*! CGAL inexact kernel with double.*/
typedef CGAL::Simple_cartesian<double> K;
/*! CGAL 3D point.*/
typedef K::Point_3 Point;
/*! CGAL 3D triangle.*/
typedef K::Triangle_3 Triangle;
/*! CGAL 3D vector.*/
typedef K::Vector_3 Vector_3;
/*! 3D plane.*/
typedef K::Plane_3 Plane;

/*! \brief Constant pi from boost library.*/
const static double pi = boost::math::constants::pi<double>();

/*!	\brief Clip a 3D vector within a sphere.

*/
inline bool sphere_clip_vector(const Point &O, double r,const Point &P, Vector_3 &V)
{
	Vector_3 W = P - O ;
    double a = V.squared_length();
	if (a == 0)
		return true;
    double b = V * W ;
    double c = W.squared_length() - r * r ;
    double delta = b * b - a * c;
    if (delta < 0)
        return true;	//numerical precision
    double t = (- b + 2 * std::sqrt(delta)) / (2 * a) ;
    if (t < 0)		
       return true;	//numerical precision
    if (t >= 1)
       return false ;	//inside sphere
	if (t == 0)		
		t = 0.01;
    V = V * t;
    return true;
}

/*! \brief Structure that contains all the parameters of the MSDM estimation

This is the C translation of the MATLAB option structure.

*/
struct MSDM_options
{	
	/*! Use multithreadding computation. Default: true.*/
	bool multithreading;
	/*! Specify number of threads to use. Default: max system - 1.*/
	int nb_threads;
	/*! Display progress information in MATLAB console. Default: true.*/
	bool display;
	/*! Use a non-connectivity dependent neighborhood definition. Default: true.*/
	bool isGeodesic;
	/*! Radius of the neighborhood. Default: 0.015.*/
	double radius;
	/*! Weight of L in the local MSDM score. Default: 0.4. */
	double alpha;
	/*! Weight of C in the local MSDM score. Default: 0.4. */
	double beta;
	/*! Weight of S in the local MSDM score. Default: 0.2. */
	double gamma;
	/*! Exponent in the Minkowski sum of local MSDM scores. Default: 3.*/
	double Mpower;
	/*! Relative radius with regard to the bounding box. Default: true.*/
	bool isRelative;
};

/*! \brief Extract a scalar value associated with a field in a MATLAB structure.
	\param MATLAB_struct Pointer to an array of MATLAB structures.
	\param index Index of the structure in the array.
	\param fieldname Name of the field in the structure.
	\param out Pointer to the scalar value associated to the fieldname.
	\tparam T Type of the scalar value.

	MATLAB scalar are represented as double. The pointer to the out value can be of any scalar type, as
	the conversion is automatically performed. If the field does not exist in the structure, the original
	value of out is unmodified.
*/
template<typename T>
inline void getMatlabFieldScalar(const mxArray* MATLAB_struct, mwIndex index, const char* fieldname, T* out)
{
	mxArray* field = mxGetField(MATLAB_struct, index, fieldname);
#pragma warning (push)
#pragma warning (disable: 4800) 
	if (field != NULL)
		*out = static_cast<T>(mxGetScalar(field)); 
#pragma warning (pop)
}

/*!	\brief Normalize a CGAL vector using according to its Euclidean norm
\param v Vector to be normalized
\return Vector with unit Euclidean norm
*/
inline Vector_3 normalize(const Vector_3& v)
{
	double l = v.squared_length();
	return (l != 0.0) ? (v / std::sqrt(l)) : v;
}

/*! \brief Returns asin or \f$\pm\frac{\pi}{2}\f$ if outside the range.*/
inline double asinT(double sine)
{
	if (sine >= 1)
		return pi/2;
	else
	{
		if (sine <= -1)
			return -pi/2;
		else
			return std::asin(sine);
	}
}


#include "enriched_polyhedron.h"
/*! CGAL polyhedron for MSDM computation.*/
typedef Enriched_polyhedron<K, Enriched_items> Polyhedron;

#endif