/*\file Distance3D.cpp
\brief Implementation of the normal cycle analysis to compute the vertex curvatures

Some unnecessary parts of the curvature computation are commented out, since they are not used in the MSDM computation.

\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re
*/

#include "normal_cycle.h"

using namespace NormalCycle;

mwSize NormalCycle::principal_curvatures(Polyhedron &pMesh, const MSDM_options& options)
{
	double radius = (options.isRelative) ? 
		(options.radius * pMesh.getMaxDim()) : 
		(options.radius);
	mwSize failures = 0;
	Polyhedron::Vertex_iterator pVertex = pMesh.vertices_begin();
	Polyhedron::Vertex_iterator pVe = pMesh.vertices_end();
	const int nb_vertices = (const int) pMesh.size_of_vertices();
	if (!options.multithreading)
	{
		if (options.isGeodesic)
		{
			CGAL_For_all(pVertex, pVe)
			{
				Eigen::Matrix3d CovMat = Eigen::Matrix3d::Zero();
				normal_cycle_patch(pVertex, CovMat, radius);
				if (!tensor2Curvatures(pVertex, CovMat, pMesh.getMaxDim()))
					failures++;
			}
		}
		else
		{
			CGAL_For_all(pVertex, pVe)
			{
				Eigen::Matrix3d CovMat = Eigen::Matrix3d::Zero();
				normal_cycle_1ring(pVertex, CovMat);
				if (!tensor2Curvatures(pVertex, CovMat))
					failures++;
			}
		}
	}
	else
	{
		std::vector<Polyhedron::Vertex_handle> vertices;
		vertices.reserve(nb_vertices);
		CGAL_For_all(pVertex, pVe){vertices.push_back(pVertex);}
		if (options.isGeodesic)
		{
#pragma omp parallel for num_threads(options.nb_threads)
			for (int i = 0 ; i < nb_vertices ; i++)
			{
				Eigen::Matrix3d CovMat = Eigen::Matrix3d::Zero();
				normal_cycle_patch(vertices[i], CovMat, radius);			
				if (!tensor2Curvatures(vertices[i], CovMat, pMesh.getMaxDim()))
#pragma omp atomic
					failures++;
			}
		}
		else
		{
#pragma omp parallel for num_threads(options.nb_threads)
			for (int i = 0 ; i < nb_vertices ; i++)
			{
				Eigen::Matrix3d CovMat = Eigen::Matrix3d::Zero();
				normal_cycle_1ring(vertices[i], CovMat);
				if (!tensor2Curvatures(vertices[i], CovMat))
#pragma omp atomic
					failures++;
			}
		}
	}
	if (options.display && failures != 0)
		mexWarnMsgIdAndTxt("MSDM_mex:normal_cyles", "could not compute all curvatures (%d failures)", failures); 	
	return failures;
}

bool NormalCycle::tensor2Curvatures(Polyhedron::Vertex_handle& pVertex, const Eigen::Matrix3d& CovMat, double radius)
{
	Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> sae(CovMat);
	if (sae.info() != Eigen::Success)
	{
		//pVertex->VKmax(CGAL::NULL_VECTOR);
		//pVertex->VKmin(CGAL::NULL_VECTOR);
		pVertex->set_kmax(0);
		pVertex->set_kmin(0);
		pVertex->set_courbureMSDM(0);
		return false;
	}
	Eigen::Array3d eVal(sae.eigenvalues());
	int id_max = 2;
	int id_min = 0;
	Eigen::Array3d eValA = eVal.abs();
	pVertex->set_kmax(eValA.maxCoeff(&id_max));
	eValA.minCoeff(&id_min);	
	int id_mean = 3 - id_max - id_min;	//min eigenvalue is not used
	/*	No eigenvector and reorientation here
		Eigen::Matrix3d eVec = sae.eigenvectors();
		pVertex->VKmax(Vector_3(eVec(0, id_max),eVec(1,id_max),eVec(2,id_max)));
		pVertex->VKmin(Vector_3(eVec(0, id_mean),eVec(1,id_mean),eVec(2,id_mean)));
		//TODO: reorient according to vertex normal
		if (eVal(id_max) > 0)
			pVertex->set_kmin(eVal(id_mean));
		else
			pVertex->set_kmin(eVal(id_mean) * (eVal(id_mean) * eVal(id_max) < 0) ? -1 : 1);
	*/
	pVertex->set_kmin(eValA(id_mean));
	pVertex->set_courbureMSDM(0.5 * radius * (pVertex->get_kmax() + pVertex->get_kmin()));
	return true;
}

inline bool NormalCycle::normal_cycle_1ring(Polyhedron::Vertex_const_handle pVertex,
	Eigen::Matrix3d& CovMat)
{
	double area = 0;
	Polyhedron::Halfedge_around_vertex_const_circulator pHalfedge = pVertex->vertex_begin();
	Polyhedron::Halfedge_around_vertex_const_circulator pHalfedgeStart = pHalfedge;
	CGAL_For_all(pHalfedge, pHalfedgeStart)
	{
		const Point& p1 = pHalfedge->vertex()->point();
		const Point& p2 = pHalfedge->opposite()->vertex()->point();
		Vector_3 edge = (p1 - p2);
		double len_edge = std::sqrt(edge.squared_length());
		if (len_edge == 0)
			continue;
		Polyhedron::Facet_const_handle pFacet1 = pHalfedge->facet();
		Polyhedron::Facet_const_handle pFacet2 = pHalfedge->opposite()->facet();
		CGAL_assertion(pFacet1 != pFacet2);
		if (pFacet1 == NULL || pFacet2 == NULL)
			continue;
		area += pHalfedge->facet()->get_area();
		const Vector_3& normal1 = pFacet1->normal();
		const Vector_3& normal2 = pFacet2->normal();
		double sine = (CGAL::cross_product(normal1, normal2) * edge) / len_edge;
		double beta = asinT(sine);
		Eigen::Vector3d Eedge(edge.x(),edge.y(),edge.z());
		CovMat += beta / len_edge * Eedge * Eedge.transpose();
	}
	if (area == 0)
		return false;
	CovMat /= area;
	return true;
}

inline bool NormalCycle::normal_cycle_patch(Polyhedron::Vertex_const_handle pVertex,
	Eigen::Matrix3d& CovMat, double radius)
{
	std::set<Polyhedron::Vertex_const_handle> vertices;
	std::stack<Polyhedron::Vertex_const_handle> S;
	vertices.insert(pVertex);
	S.push(pVertex);
	const Point& O = pVertex->point();
	int iter = 0;
	while(!S.empty())
	{
		Polyhedron::Vertex_const_handle v = S.top();
		S.pop();
		const Point& P = v->point();
		Polyhedron::Halfedge_around_vertex_const_circulator h = v->vertex_begin();
		Polyhedron::Halfedge_around_vertex_const_circulator pHalfedgeStart = h;
		CGAL_For_all(h, pHalfedgeStart)
		{
			const Point& p1 = h->vertex()->point();
			const Point& p2 = h->opposite()->vertex()->point();
			Vector_3 V = (p2 - p1);
			if (v == pVertex || V * (P - O) > 0.0) 
			{
				bool isect = sphere_clip_vector(O, radius, P, V) ;
				if (!h->is_border_edge()) 
				{
					double len_edge = std::sqrt(V.squared_length());
					Polyhedron::Facet_const_handle pFacet1 = h->facet();
					Polyhedron::Facet_const_handle pFacet2 = h->opposite()->facet();
					CGAL_assertion(pFacet1 != pFacet2);
					if (pFacet1 == NULL || pFacet2 == NULL)
						continue;
					const Vector_3& normal1 = pFacet1->normal();
					const Vector_3& normal2 = pFacet2->normal();
					double sine = (CGAL::cross_product(normal1, normal2) * V) / len_edge;
					double beta = asinT(sine);
					Eigen::Vector3d edge(V.x(),V.y(),V.z());
					CovMat += beta / len_edge * edge * edge.transpose();
				}
				if (!isect)
				{
					Polyhedron::Vertex_const_iterator w=h->opposite()->vertex();
					if(vertices.find(w) == vertices.end()) 
					{
						vertices.insert(w) ;
						S.push(w) ;
					}
				}
			}
		}
		iter++;
	}
	double area = pi * radius * radius;
	if (area == 0)
		return false;
	CovMat /= area;
	return true;
}