#ifndef NORMAL_CYCLE_H
#define NORMAL_CYCLE_H

#include "Header.h"
#include <set>
#include <stack>
#include <Eigen/Eigenvalues> 

/*! \brief Estimation of principale curvatures based on the normal cycle
\copyright LIRIS, M2DisCo 2009, Licence QPL
\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re

According to: <i>Restricted Delaunay Triangulations and Normal Cycle</i>, 
David Cohen-Steiner and J.M. Morvan In Proceedings SoCG'03.
*/
namespace NormalCycle
{
	/*! \brief principal curvature for all vertices.
		\param pMesh Mesh to be analyzed.
		\param options Parameter settings for the curvature analysis.
		\return Number of failures in the curvature estimation process
		

		This function is procedure is multithreaded. Curvatures are estimated using the normal
		cycle, either integrated on the 1-ring neighborhood, or a local Euclidean-based neighborhood.
	*/
	mwSize principal_curvatures(Polyhedron &pMesh, const MSDM_options& options);

	/*! \brief tensor for a vertex using 1-ring neighborhood.
		\param pVertex Vertex to be analyzed.
		\param[out] CovMat integral of the curvature tensor.
		\return true if computation succeeds.
		
		Compute the normal cycle on the 1 ring neighborhood around a vertex.
	*/
	inline bool normal_cycle_1ring(Polyhedron::Vertex_const_handle pVertex, Eigen::Matrix3d& CovMat);

	/*! \brief tensor for a vertex using patch neighborhood.
		\param pVertex Vertex to be analyzed.
		\param[out] CovMat integral of the curvature tensor.
		\param radius Radius of the neighborhood patch
		\return true if computation succeeds.
		
		Compute the normal cycle on a surface patch neighborhood around a vertex.

		\sa sphere_clip_vector
	*/
	inline bool normal_cycle_patch(Polyhedron::Vertex_const_handle pVertex,
		Eigen::Matrix3d& CovMat, double radius);

	/*! \brief Estimate curvatures from tensor.
		\param[out] pVertex Vertex whose curvatures are to be computed
		\param CovMat Curvature tensor
		\param radius Scale mean curvature
		\return true if computation succeeds.

		The estimated curvature tensor around the vertex query is eigendecomposed. The eigenvalues and eigenvectors
		correspond to the principal curvatures of the vertex, which are updated accordingly.

		The MSDM uses a scaled mean curvature at each vertex, which is computed here.

		<b>Some parts of the principal curvature computation that are not needed in the MSDM are skipped.</b>
		More precisely:
		- The computation of the principal direction is not performed.
		- The curvatures stay unsigned (not oriented w.r.t. the normal).
		
		The MSDM indeed uses \f$0.5\times(|\kappa_{\min}|+\kappa_{\max})\times \mathrm{radius}\f$ in 
		the distance computation.
	*/
	inline bool tensor2Curvatures(Polyhedron::Vertex_handle& pVertex, const Eigen::Matrix3d& CovMat, double radius = 1);
}

#endif